@@ -41,6 +41,10 @@ grn_obj *mrn_db;
 grn_hash *mrn_hash;
 pthread_mutex_t db_mutex;
 
+/* status */
+st_mrn_statuses mrn_status_vals;
+long mrn_count_skip = 0;
+
 /* logging */
 const char *mrn_logfile_name = MRN_LOG_FILE_NAME;
 FILE *mrn_logfile = NULL;
@@ -66,12 +70,35 @@ grn_logger_info mrn_logger_info = {
 
 /* system functions */
 
+static void mrn_create_status()
+{
+  DBUG_ENTER("mrn_create_status");
+  mrn_status_vals.count_skip = mrn_count_skip;
+  DBUG_VOID_RETURN;
+}
+
+struct st_mysql_show_var mrn_statuses[] =
+{
+  {"count_skip", (char *) &mrn_status_vals.count_skip, SHOW_LONG},
+  {NullS, NullS, SHOW_LONG}
+};
+
+static int mrn_show_status(THD *thd, SHOW_VAR *var, char *buff)
+{
+  DBUG_ENTER("mrn_show_status");
+  mrn_create_status();
+  var->type = SHOW_ARRAY;
+  var->value = (char *) &mrn_statuses;
+  DBUG_RETURN(0);
+}
+
 struct st_mysql_storage_engine storage_engine_structure =
 { MYSQL_HANDLERTON_INTERFACE_VERSION };
 
 struct st_mysql_show_var mrn_status_variables[] =
 {
-  {NULL}
+  {"groonga", (char *) &mrn_show_status, SHOW_FUNC},
+  {NullS, NullS, SHOW_LONG}
 };
 
 struct st_mysql_sys_var *mrn_system_variables[] =
@@ -581,6 +608,7 @@ ha_mroonga::ha_mroonga(handlerton *hton, TABLE_SHARE *share)
   DBUG_ENTER("ha_mroonga::ha_mroonga");
   ctx = grn_ctx_open(0);
   grn_ctx_use(ctx, mrn_db);
+  cur = NULL;
   DBUG_VOID_RETURN;
 }
 
@@ -625,7 +653,7 @@ ulonglong ha_mroonga_table_flags =
     HA_REC_NOT_IN_SEQ |
     HA_NULL_IN_KEY |
     HA_CAN_INDEX_BLOBS |
-    //HA_STATS_RECORDS_IS_EXACT |
+    HA_STATS_RECORDS_IS_EXACT |
     HA_NO_PREFIX_CHAR_KEYS |
     HA_CAN_FULLTEXT |
     HA_NO_AUTO_INCREMENT |
@@ -676,10 +704,11 @@ int ha_mroonga::create(const char *name, TABLE *table, HA_CREATE_INFO *info)
     } else {
       // opening existing database
       db_obj = grn_db_open(ctx, db_path);
-      if (db_obj == NULL) {
+      if (ctx->rc) {
         GRN_LOG(ctx, GRN_LOG_ERROR, "cannot open database (%s)", db_path);
+        my_message(ER_CANT_OPEN_FILE, "cannot open database file", MYF(0));
         pthread_mutex_unlock(&db_mutex);
-        DBUG_RETURN(-1);
+        DBUG_RETURN(ER_CANT_OPEN_FILE);
       }
     }
     mrn_hash_put(ctx, mrn_hash, db_name, db_obj);
@@ -731,10 +760,10 @@ int ha_mroonga::create(const char *name, TABLE *table, HA_CREATE_INFO *info)
 
   tbl_obj = grn_table_create(ctx, tbl_name, tbl_name_len, tbl_path,
                          tbl_flags, pkey_type, pkey_value_type);
-
-  if (tbl_obj == NULL) {
+  if (ctx->rc) {
     GRN_LOG(ctx, GRN_LOG_ERROR, "cannot create table: name=%s", tbl_name);
-    DBUG_RETURN(-1);
+    my_message(ER_CANT_CREATE_TABLE, ctx->errbuf, MYF(0));
+    DBUG_RETURN(ER_CANT_CREATE_TABLE);
   }
 
   /* create columns */
@@ -753,11 +782,12 @@ int ha_mroonga::create(const char *name, TABLE *table, HA_CREATE_INFO *info)
 
     col_obj = grn_column_create(ctx, tbl_obj, col_name, col_name_size,
                                 col_path, col_flags, col_type);
-    if (col_obj == NULL) {
+    if (ctx->rc) {
       GRN_LOG(ctx, GRN_LOG_ERROR, "cannot create table: name=%s, col=%s",
               tbl_name, col_name);
       grn_obj_remove(ctx, tbl_obj);
-      DBUG_RETURN(-1);
+      my_message(ER_CANT_CREATE_TABLE, ctx->errbuf, MYF(0));
+      DBUG_RETURN(ER_CANT_CREATE_TABLE);
     }
   }
 
@@ -795,19 +825,20 @@ int ha_mroonga::create(const char *name, TABLE *table, HA_CREATE_INFO *info)
     int key_alg = key_info.algorithm;
     grn_obj_flags idx_tbl_flags;
     if (key_alg == HA_KEY_ALG_FULLTEXT) {
-      idx_tbl_flags = GRN_OBJ_TABLE_PAT_KEY | GRN_OBJ_PERSISTENT;
+      idx_tbl_flags = GRN_OBJ_TABLE_PAT_KEY | GRN_OBJ_PERSISTENT | GRN_OBJ_KEY_NORMALIZE;
     } else if (key_alg == HA_KEY_ALG_HASH) {
-      idx_tbl_flags = GRN_OBJ_TABLE_HASH_KEY | GRN_OBJ_PERSISTENT;
+      idx_tbl_flags = GRN_OBJ_TABLE_HASH_KEY | GRN_OBJ_PERSISTENT | GRN_OBJ_KEY_NORMALIZE;
     } else {
-      idx_tbl_flags = GRN_OBJ_TABLE_PAT_KEY | GRN_OBJ_PERSISTENT;
+      idx_tbl_flags = GRN_OBJ_TABLE_PAT_KEY | GRN_OBJ_PERSISTENT | GRN_OBJ_KEY_NORMALIZE;
     }
 
     idx_tbl_obj = grn_table_create(ctx, idx_name, strlen(idx_name), NULL,
                                    idx_tbl_flags, col_type, 0);
-    if (idx_tbl_obj == NULL) {
+    if (ctx->rc) {
       GRN_LOG(ctx, GRN_LOG_ERROR, "cannot create index: name=%s", idx_name);
       grn_obj_remove(ctx, tbl_obj);
-      DBUG_RETURN(-1);
+      my_message(ER_CANT_CREATE_TABLE, ctx->errbuf, MYF(0));
+      DBUG_RETURN(ER_CANT_CREATE_TABLE);
     }
 
     if (key_alg == HA_KEY_ALG_FULLTEXT) {
@@ -824,7 +855,8 @@ int ha_mroonga::create(const char *name, TABLE *table, HA_CREATE_INFO *info)
               idx_name, col_name);
       grn_obj_remove(ctx, idx_tbl_obj);
       grn_obj_remove(ctx, tbl_obj);
-      DBUG_RETURN(-1);
+      my_message(ER_CANT_CREATE_TABLE, ctx->errbuf, MYF(0));
+      DBUG_RETURN(ER_CANT_CREATE_TABLE);
     }
 
     grn_id gid = grn_obj_id(ctx, col_obj);
@@ -857,10 +889,10 @@ int ha_mroonga::open(const char *name, int mode, uint test_if_locked)
   // we should not call grn_db_open() very often. so we use cache.
   if (mrn_hash_get(ctx, mrn_hash, db_name, (void**) &(db)) != 0) {
     db = grn_db_open(ctx, db_path);
-    if (db == NULL) {
+    if (ctx->rc) {
       GRN_LOG(ctx, GRN_LOG_ERROR, "cannot open database (%s)", db_path);
       pthread_mutex_unlock(&db_mutex);
-      DBUG_RETURN(-1);
+      DBUG_RETURN(ER_CANT_OPEN_FILE);
     }
     mrn_hash_put(ctx, mrn_hash, db_name, db);
   }
@@ -871,9 +903,9 @@ int ha_mroonga::open(const char *name, int mode, uint test_if_locked)
   char tbl_name[MRN_MAX_PATH_SIZE];
   mrn_table_name_gen(name, tbl_name);
   tbl = grn_ctx_get(ctx, tbl_name, strlen(tbl_name));
-  if (tbl == NULL) {
+  if (ctx->rc) {
     GRN_LOG(ctx, GRN_LOG_ERROR, "cannot open table (%s)", tbl_name);
-    DBUG_RETURN(-1);
+    DBUG_RETURN(ER_CANT_OPEN_FILE);
   }
 
   /* open columns */
@@ -886,11 +918,11 @@ int ha_mroonga::open(const char *name, int mode, uint test_if_locked)
     const char *col_name = field->field_name;
     int col_name_size = strlen(col_name);
     col[i] = grn_obj_column(ctx, tbl, col_name, col_name_size);
-    if (col[i] == NULL) {
+    if (ctx->rc) {
       GRN_LOG(ctx, GRN_LOG_ERROR, "cannot open table(col) %s(%s)",
               tbl_name, col_name);
       grn_obj_unlink(ctx, tbl);
-      DBUG_RETURN(-1);
+      DBUG_RETURN(ER_CANT_OPEN_FILE);
     }
   }
 
@@ -919,12 +951,25 @@ int ha_mroonga::open(const char *name, int mode, uint test_if_locked)
 
     mrn_index_name_gen(tbl_name, i, idx_name);
     idx_tbl[i] = grn_ctx_get(ctx, idx_name, strlen(idx_name));
+    if (ctx->rc) {
+      GRN_LOG(ctx, GRN_LOG_ERROR, "cannot open table(index) %s(%s)",
+              tbl_name, idx_name);
+      grn_obj_unlink(ctx, tbl);
+      DBUG_RETURN(ER_CANT_OPEN_FILE);
+    }
 
     KEY key_info = table->s->key_info[i];
     Field *field = key_info.key_part[0].field;
     const char *col_name = field->field_name;
     int col_name_size = strlen(col_name);
     idx_col[i] = grn_obj_column(ctx, idx_tbl[i], col_name, col_name_size);
+    if (ctx->rc) {
+      GRN_LOG(ctx, GRN_LOG_ERROR, "cannot open index(col) %s(%s)",
+              idx_name, col_name);
+      grn_obj_unlink(ctx, idx_tbl[i]);
+      grn_obj_unlink(ctx, tbl);
+      DBUG_RETURN(ER_CANT_OPEN_FILE);
+    }
   }
   DBUG_RETURN(0);
 }
@@ -1016,6 +1061,7 @@ THR_LOCK_DATA **ha_mroonga::store_lock(THD *thd, THR_LOCK_DATA **to,
 int ha_mroonga::rnd_init(bool scan)
 {
   DBUG_ENTER("ha_mroonga::rnd_init");
+  count_skip = FALSE;
   cur = grn_table_cursor_open(ctx, tbl, NULL, 0, NULL, 0, 0, -1, 0);
   if (cur == NULL) {
       GRN_LOG(ctx, GRN_LOG_ERROR, "cannot open cursor");
@@ -1024,21 +1070,28 @@ int ha_mroonga::rnd_init(bool scan)
   DBUG_RETURN(0);
 }
 
+int ha_mroonga::rnd_end()
+{
+  DBUG_ENTER("ha_mroonga::rnd_end");
+  if (cur) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+  }
+  DBUG_RETURN(0);
+}
+
 int ha_mroonga::rnd_next(uchar *buf)
 {
   DBUG_ENTER("ha_mroonga::rnd_next");
   row_id = grn_table_cursor_next(ctx, cur);
   if (row_id == GRN_ID_NIL) {
     grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
     DBUG_RETURN(HA_ERR_END_OF_FILE);
   }
-  int i;
-  int n_columns = table->s->fields;
-  for (i = 0; i < n_columns; i++) {
-    Field *field = table->field[i];
-    bitmap_set_bit(table->write_set, field->field_index);
-    mrn_store_field(ctx, field, col[i], row_id);
-  }
+  store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
   DBUG_RETURN(0);
 }
 
@@ -1046,13 +1099,7 @@ int ha_mroonga::rnd_pos(uchar *buf, uchar *pos)
 {
   DBUG_ENTER("ha_mroonga::rnd_pos");
   row_id = *((grn_id*) pos);
-  int i;
-  int n_columns = table->s->fields;
-  for (i = 0; i < n_columns; i++) {
-    Field *field = table->field[i];
-    bitmap_set_bit(table->write_set, field->field_index);
-    mrn_store_field(ctx, field, col[i], row_id);
-  }
+  store_fields_from_primary_table(buf, row_id);
   DBUG_RETURN(0);
 }
 
@@ -1071,12 +1118,14 @@ int ha_mroonga::write_row(uchar *buf)
   int pkey_size = 0;
   uint pkeynr = table->s->primary_key;
   GRN_VOID_INIT(&wrapper);
+#ifndef DBUG_OFF
+  my_bitmap_map *tmp_map = dbug_tmp_use_all_columns(table, table->read_set);
+#endif
   if (pkeynr != MAX_INDEXES) {
     KEY key_info = table->s->key_info[pkeynr];
     // surpose simgle column key
     int field_no = key_info.key_part[0].field->field_index;
     Field *pkey_field = table->field[field_no];
-    bitmap_set_bit(table->read_set, pkey_field->field_index);
     mrn_set_buf(ctx, pkey_field, &wrapper, &pkey_size);
     pkey = GRN_TEXT_VALUE(&wrapper);
   }
@@ -1086,6 +1135,9 @@ int ha_mroonga::write_row(uchar *buf)
   grn_obj_unlink(ctx, &wrapper);
   if (added == 0) {
     // duplicated error
+#ifndef DBUG_OFF
+    dbug_tmp_restore_column_map(table->read_set, tmp_map);
+#endif
     DBUG_RETURN(-1);
   }
 
@@ -1095,14 +1147,19 @@ int ha_mroonga::write_row(uchar *buf)
   GRN_VOID_INIT(&colbuf);
   for (i = 0; i < n_columns; i++) {
     Field *field = table->field[i];
-    bitmap_set_bit(table->read_set, field->field_index);
     mrn_set_buf(ctx, field, &colbuf, &col_size);
     if (grn_obj_set_value(ctx, col[i], row_id, &colbuf, GRN_OBJ_SET)
         != GRN_SUCCESS) {
+#ifndef DBUG_OFF
+      dbug_tmp_restore_column_map(table->read_set, tmp_map);
+#endif
       grn_obj_unlink(ctx, &colbuf);
       DBUG_RETURN(-1);
     }
   }
+#ifndef DBUG_OFF
+  dbug_tmp_restore_column_map(table->read_set, tmp_map);
+#endif
   grn_obj_unlink(ctx, &colbuf);
   DBUG_RETURN(0);
 }
@@ -1116,12 +1173,24 @@ int ha_mroonga::update_row(const uchar *old_data, uchar *new_data)
   GRN_VOID_INIT(&colbuf);
   for (i = 0; i < n_columns; i++) {
     Field *field = table->field[i];
-    bitmap_set_bit(table->read_set, field->field_index);
-    mrn_set_buf(ctx, field, &colbuf, &col_size);
-    if (grn_obj_set_value(ctx, col[i], row_id, &colbuf, GRN_OBJ_SET)
-        != GRN_SUCCESS) {
-      grn_obj_unlink(ctx, &colbuf);
-      DBUG_RETURN(-1);
+    if (bitmap_is_set(table->write_set, field->field_index)) {
+#ifndef DBUG_OFF
+      my_bitmap_map *tmp_map = dbug_tmp_use_all_columns(table,
+        table->read_set);
+#endif
+      DBUG_PRINT("info",("mroonga update column %d(%d)",i,field->field_index));
+      mrn_set_buf(ctx, field, &colbuf, &col_size);
+      if (grn_obj_set_value(ctx, col[i], row_id, &colbuf, GRN_OBJ_SET)
+          != GRN_SUCCESS) {
+#ifndef DBUG_OFF
+        dbug_tmp_restore_column_map(table->read_set, tmp_map);
+#endif
+        grn_obj_unlink(ctx, &colbuf);
+        DBUG_RETURN(-1);
+      }
+#ifndef DBUG_OFF
+      dbug_tmp_restore_column_map(table->read_set, tmp_map);
+#endif
     }
   }
   grn_obj_unlink(ctx, &colbuf);
@@ -1150,14 +1219,14 @@ ha_rows ha_mroonga::records_in_range(uint keynr, key_range *range_min, key_range
   if (range_min != NULL) {
     mrn_set_key_buf(ctx, field, range_min->key, key_min[keynr], &size_min);
     val_min = key_min[keynr];
-    if (range_min->flag & HA_READ_AFTER_KEY) {
+    if (range_min->flag == HA_READ_AFTER_KEY) {
       flags |= GRN_CURSOR_GT;
     }
   }
   if (range_max != NULL) {
     mrn_set_key_buf(ctx, field, range_max->key, key_max[keynr], &size_max);
     val_max = key_max[keynr];
-    if (range_max->flag & HA_READ_BEFORE_KEY) {
+    if (range_max->flag == HA_READ_BEFORE_KEY) {
       flags |= GRN_CURSOR_LT;
     }
   }
@@ -1186,39 +1255,136 @@ ha_rows ha_mroonga::records_in_range(uint keynr, key_range *range_min, key_range
   DBUG_RETURN(row_count);
 }
 
-int ha_mroonga::index_read(uchar * record_buffer, const uchar * key, uint key_len,
-                           enum ha_rkey_function find_flag)
+int ha_mroonga::index_init(uint idx, bool sorted)
 {
-  DBUG_ENTER("ha_mroonga::index_read");
-  uint keynr = active_index;
-  uint pkeynr = table->s->primary_key;
-  KEY key_info = table->key_info[keynr];
-  KEY_PART_INFO key_part = key_info.key_part[0];
-  if (keynr == pkeynr) {
-    row_id = grn_table_get(ctx, tbl, key, key_len);
-    int i;
-    int n_columns = table->s->fields;
-    for (i = 0; i < n_columns; i++) {
-      Field *field = table->field[i];
-      if (bitmap_is_set(table->read_set, field->field_index)) {
-        bitmap_set_bit(table->write_set, field->field_index);
-        mrn_store_field(ctx, field, col[i], row_id);
-      }
-    }
+  DBUG_ENTER("ha_mroonga::index_init");
+  active_index = idx;
+  count_skip = FALSE;
+  DBUG_RETURN(0);
+}
+
+int ha_mroonga::index_end()
+{
+  DBUG_ENTER("ha_mroonga::index_end");
+  if (cur) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
   }
   DBUG_RETURN(0);
 }
 
-int ha_mroonga::index_read_idx(uchar * buf, uint index, const uchar * key,
-                               uint key_len, enum ha_rkey_function find_flag)
+int ha_mroonga::index_read_map(uchar * buf, const uchar * key,
+                               key_part_map keypart_map,
+                               enum ha_rkey_function find_flag)
 {
-  DBUG_ENTER("ha_mroonga::index_read_idx");
+  DBUG_ENTER("ha_mroonga::index_read_map");
+  uint keynr = active_index;
+  KEY key_info = table->key_info[keynr];
+  KEY_PART_INFO key_part = key_info.key_part[0];
+  check_count_skip(keypart_map, 0, FALSE);
+
+  int flags = 0;
+  uint size_min = 0, size_max = 0;
+  void *val_min = NULL, *val_max = NULL;
+  Field *field = key_part.field;
+
+  if (cur) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+  }
+
+  if (find_flag == HA_READ_KEY_EXACT) {
+    mrn_set_key_buf(ctx, field, key, key_min[keynr], &size_min);
+    val_min = key_min[keynr];
+    val_max = key_min[keynr];
+    size_max = size_min;
+  } else if (
+    find_flag == HA_READ_BEFORE_KEY ||
+    find_flag == HA_READ_PREFIX_LAST_OR_PREV
+  ) {
+    mrn_set_key_buf(ctx, field, key, key_max[keynr], &size_max);
+    val_max = key_max[keynr];
+    if (find_flag == HA_READ_BEFORE_KEY) {
+      flags |= GRN_CURSOR_LT;
+    }
+  } else {
+    mrn_set_key_buf(ctx, field, key, key_min[keynr], &size_min);
+    val_min = key_min[keynr];
+    if (find_flag == HA_READ_AFTER_KEY) {
+      flags |= GRN_CURSOR_GT;
+    }
+  }
+
+  uint pkeynr = table->s->primary_key;
+
+  if (keynr == pkeynr) { // primary index
+    DBUG_PRINT("info",("mroonga use primary key"));
+    cur =
+      grn_table_cursor_open(ctx, tbl, val_min, size_min, val_max, size_max,
+                            0, -1, flags);
+  } else { // normal index
+    DBUG_PRINT("info",("mroonga use key%u", keynr));
+    cur =
+      grn_table_cursor_open(ctx, idx_tbl[keynr], val_min, size_min,
+                            val_max, size_max, 0, -1, flags);
+  }
+  row_id = grn_table_cursor_next(ctx, cur);
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+  store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
   DBUG_RETURN(0);
 }
 
-int ha_mroonga::index_read_last(uchar *buf, const uchar *key, uint key_len)
+int ha_mroonga::index_read_last_map(uchar *buf, const uchar *key,
+                                    key_part_map keypart_map)
 {
-  DBUG_ENTER("ha_mroonga::index_read_last");
+  DBUG_ENTER("ha_mroonga::index_read_last_map");
+  uint keynr = active_index;
+  KEY key_info = table->key_info[keynr];
+  KEY_PART_INFO key_part = key_info.key_part[0];
+
+  int flags = GRN_CURSOR_DESCENDING;
+  uint size_min = 0, size_max = 0;
+  void *val_min = NULL, *val_max = NULL;
+  Field *field = key_part.field;
+
+  if (cur) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+  }
+
+  mrn_set_key_buf(ctx, field, key, key_min[keynr], &size_min);
+  val_min = key_min[keynr];
+  val_max = key_min[keynr];
+  size_max = size_min;
+
+  uint pkeynr = table->s->primary_key;
+
+  if (keynr == pkeynr) { // primary index
+    DBUG_PRINT("info",("mroonga use primary key"));
+    cur =
+      grn_table_cursor_open(ctx, tbl, val_min, size_min, val_max, size_max,
+                            0, -1, flags);
+  } else { // normal index
+    DBUG_PRINT("info",("mroonga use key%u", keynr));
+    cur =
+      grn_table_cursor_open(ctx, idx_tbl[keynr], val_min, size_min,
+                            val_max, size_max, 0, -1, flags);
+  }
+  row_id = grn_table_cursor_next(ctx, cur);
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+  store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
   DBUG_RETURN(0);
 }
 
@@ -1228,34 +1394,185 @@ int ha_mroonga::index_next(uchar *buf)
   row_id = grn_table_cursor_next(ctx, cur);
   if (row_id == GRN_ID_NIL) {
     grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
     DBUG_RETURN(HA_ERR_END_OF_FILE);
   }
-  int i;
-  int n_columns = table->s->fields;
-  for (i = 0; i < n_columns; i++) {
-    Field *field = table->field[i];
-    bitmap_set_bit(table->write_set, field->field_index);
-    mrn_store_field(ctx, field, col[i], row_id);
-  }
+  store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
   DBUG_RETURN(0);
 }
 
 int ha_mroonga::index_prev(uchar *buf)
 {
   DBUG_ENTER("ha_mroonga::index_prev");
-  DBUG_RETURN(HA_ERR_END_OF_FILE);
+  row_id = grn_table_cursor_next(ctx, cur);
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+  store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
+  DBUG_RETURN(0);
 }
 
 int ha_mroonga::index_first(uchar *buf)
 {
   DBUG_ENTER("ha_mroonga::index_first");
-  DBUG_RETURN(HA_ERR_END_OF_FILE);
+  if (cur) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+  }
+  uint pkeynr = table->s->primary_key;
+  if (active_index == pkeynr) { // primary index
+    DBUG_PRINT("info",("mroonga use primary key"));
+    cur =
+      grn_table_cursor_open(ctx, tbl, NULL, 0, NULL, 0,
+                            0, -1, 0);
+  } else { // normal index
+    DBUG_PRINT("info",("mroonga use key%u", active_index));
+    cur =
+      grn_table_cursor_open(ctx, idx_tbl[active_index], NULL, 0,
+                            NULL, 0, 0, -1, 0);
+  }
+  row_id = grn_table_cursor_next(ctx, cur);
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+  store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
+  DBUG_RETURN(0);
 }
 
 int ha_mroonga::index_last(uchar *buf)
 {
   DBUG_ENTER("ha_mroonga::index_last");
-  DBUG_RETURN(HA_ERR_END_OF_FILE);
+  if (cur) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+  }
+  int flags = GRN_CURSOR_DESCENDING;
+  uint pkeynr = table->s->primary_key;
+  if (active_index == pkeynr) { // primary index
+    DBUG_PRINT("info",("mroonga use primary key"));
+    cur =
+      grn_table_cursor_open(ctx, tbl, NULL, 0, NULL, 0,
+                            0, -1, flags);
+  } else { // normal index
+    DBUG_PRINT("info",("mroonga use key%u", active_index));
+    cur =
+      grn_table_cursor_open(ctx, idx_tbl[active_index], NULL, 0,
+                            NULL, 0, 0, -1, flags);
+  }
+  row_id = grn_table_cursor_next(ctx, cur);
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+  store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
+  DBUG_RETURN(0);
+}
+
+int ha_mroonga::index_next_same(uchar *buf, const uchar *key, uint keylen)
+{
+  DBUG_ENTER("ha_mroonga::index_next_same");
+  row_id = grn_table_cursor_next(ctx, cur);
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+  if (!count_skip)
+    store_fields_from_primary_table(buf, row_id);
+  table->status = 0;
+  DBUG_RETURN(0);
+}
+
+int ha_mroonga::read_range_first(const key_range *start_key,
+                                 const key_range *end_key,
+                                 bool eq_range, bool sorted)
+{
+  DBUG_ENTER("ha_mroonga::read_range_first");
+  check_count_skip(start_key ? start_key->keypart_map : 0,
+                   end_key ? end_key->keypart_map : 0, FALSE);
+  int flags = 0;
+  uint size_min = 0, size_max = 0;
+  void *val_min = NULL, *val_max = NULL;
+  KEY key_info = table->s->key_info[active_index];
+  KEY_PART_INFO key_part = key_info.key_part[0];
+  Field *field = key_part.field;
+
+  if (cur) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+  }
+
+  if (start_key != NULL) {
+    mrn_set_key_buf(ctx, field, start_key->key, key_min[active_index],
+                    &size_min);
+    val_min = key_min[active_index];
+    if (start_key->flag == HA_READ_AFTER_KEY) {
+      flags |= GRN_CURSOR_GT;
+    }
+  }
+  if (end_key != NULL) {
+    mrn_set_key_buf(ctx, field, end_key->key, key_max[active_index],
+      &size_max);
+    val_max = key_max[active_index];
+    if (end_key->flag == HA_READ_BEFORE_KEY) {
+      flags |= GRN_CURSOR_LT;
+    }
+  }
+  uint pkeynr = table->s->primary_key;
+
+  if (active_index == pkeynr) { // primary index
+    DBUG_PRINT("info",("mroonga use primary key"));
+    cur =
+      grn_table_cursor_open(ctx, tbl, val_min, size_min, val_max, size_max,
+                            0, -1, flags);
+  } else { // normal index
+    DBUG_PRINT("info",("mroonga use key%u", active_index));
+    cur =
+      grn_table_cursor_open(ctx, idx_tbl[active_index], val_min, size_min,
+                            val_max, size_max, 0, -1, flags);
+  }
+  row_id = grn_table_cursor_next(ctx, cur);
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+  store_fields_from_primary_table(table->record[0], row_id);
+  table->status = 0;
+  DBUG_RETURN(0);
+}
+
+int ha_mroonga::read_range_next() {
+  DBUG_ENTER("ha_mroonga::read_range_next");
+  row_id = grn_table_cursor_next(ctx, cur);
+
+  if (row_id == GRN_ID_NIL) {
+    grn_table_cursor_close(ctx, cur);
+    cur = NULL;
+    table->status = STATUS_NOT_FOUND;
+    DBUG_RETURN(HA_ERR_END_OF_FILE);
+  }
+
+  uint pkeynr = table->s->primary_key;
+  if (!count_skip)
+    store_fields_from_primary_table(table->record[0], row_id);
+  table->status = 0;
+  DBUG_RETURN(0);
 }
 
 int ha_mroonga::ft_init() {
@@ -1269,6 +1586,8 @@ FT_INFO *ha_mroonga::ft_init_ext(uint flags, uint keynr, String *key)
   grn_obj *ft = idx_col[keynr];
   const char *keyword = key->ptr();
   int keyword_size = strlen(keyword);
+  check_count_skip(0, 0, TRUE);
+  row_id = GRN_ID_NIL;
 
   res = grn_table_create(ctx, NULL, 0, NULL, GRN_TABLE_HASH_KEY, tbl, 0);
 
@@ -1298,25 +1617,19 @@ int ha_mroonga::ft_read(uchar *buf)
 
   if (rid == GRN_ID_NIL) {
     grn_table_cursor_close(ctx, cur);
+    cur = NULL;
     grn_obj_unlink(ctx, res);
+    table->status = STATUS_NOT_FOUND;
     DBUG_RETURN(HA_ERR_END_OF_FILE);
   }
+  table->status = 0;
 
-  if (table->in_use->lex->select_lex.join->tmp_table_param.sum_func_count > 0) {
+  if (count_skip && row_id != GRN_ID_NIL) {
     DBUG_RETURN(0);
   }
 
   grn_table_get_key(ctx, res, rid, &row_id, sizeof(grn_id));
-
-  int i;
-  int n_columns = table->s->fields;
-  for (i = 0; i < n_columns; i++) {
-    Field *field = table->field[i];
-    if (bitmap_is_set(table->read_set, field->field_index)) {
-      bitmap_set_bit(table->write_set, field->field_index);
-      mrn_store_field(ctx, field, col[i], row_id);
-    }
-  }
+  store_fields_from_primary_table(buf, row_id);
   DBUG_RETURN(0);
 }
 
@@ -1332,6 +1645,130 @@ void ha_mroonga::cond_pop()
   DBUG_VOID_RETURN;
 }
 
+bool ha_mroonga::get_error_message(int error, String *buf)
+{
+  DBUG_ENTER("ha_mroonga::get_error_message");
+  // latest error message
+  buf->copy(ctx->errbuf, (uint) strlen(ctx->errbuf), system_charset_info);
+  DBUG_RETURN(FALSE);
+}
+
+void ha_mroonga::check_count_skip(key_part_map start_key_part_map,
+                                  key_part_map end_key_part_map, bool fulltext)
+{
+  DBUG_ENTER("ha_mroonga::check_count_skip");
+  st_select_lex	*select_lex = table->pos_in_table_list->select_lex;
+
+  if (
+    thd_sql_command(ha_thd()) == SQLCOM_SELECT &&
+    !select_lex->non_agg_fields.elements &&
+    !select_lex->group_list.elements &&
+    !select_lex->having &&
+    select_lex->table_list.elements == 1
+  ) {
+    Item *info = (Item *) select_lex->item_list.first_node()->info;
+    if (
+      info->type() != Item::SUM_FUNC_ITEM ||
+      ((Item_sum *) info)->sum_func() != Item_sum::COUNT_FUNC ||
+      ((Item_sum *) info)->nest_level ||
+      ((Item_sum *) info)->aggr_level ||
+      ((Item_sum *) info)->max_arg_level != -1 ||
+      ((Item_sum *) info)->max_sum_func_level != -1
+    ) {
+      count_skip = FALSE;
+      DBUG_VOID_RETURN;
+    }
+
+    int i = 0;
+    Item *where;
+    if (fulltext) {
+      where = select_lex->where;
+      if (!where ||
+          where->type() != Item::FUNC_ITEM ||
+          ((Item_func *)where)->functype() != Item_func::FT_FUNC) {
+        count_skip = FALSE;
+        DBUG_VOID_RETURN;
+      }
+      where = where->next;
+      if (!where ||
+          where->type() != Item::STRING_ITEM) {
+        count_skip = FALSE;
+        DBUG_VOID_RETURN;
+      }
+      for (where = where->next; where; where = where->next) {
+        if (where->type() != Item::FIELD_ITEM)
+          break;
+      }
+      if (where != info) {
+        count_skip = FALSE;
+        DBUG_VOID_RETURN;
+      }
+      count_skip = TRUE;
+      mrn_count_skip++;
+      DBUG_VOID_RETURN;
+    } else {
+      uint keynr = active_index;
+      KEY key_info = table->key_info[keynr];
+      KEY_PART_INFO *key_part = key_info.key_part;
+      for (where = select_lex->where; where; where = where->next) {
+        if (where->type() == Item::FIELD_ITEM)
+        {
+          Field *field = ((Item_field *)where)->field;
+          if (field->table != table)
+            break;
+          int j;
+          for (j = 0; j < key_info.key_parts; j++) {
+            if (key_part[j].field == field)
+            {
+              if (!(start_key_part_map >> j) && !(end_key_part_map >> j))
+                j = key_info.key_parts;
+              else
+                i++;
+              break;
+            }
+          }
+          if (j >= key_info.key_parts)
+            break;
+        }
+        if (i >= select_lex->select_n_where_fields)
+        {
+          count_skip = TRUE;
+          mrn_count_skip++;
+          DBUG_VOID_RETURN;
+        }
+      }
+    }
+  }
+  count_skip = FALSE;
+  DBUG_VOID_RETURN;
+}
+
+void ha_mroonga::store_fields_from_primary_table(uchar *buf, grn_id rid)
+{
+  DBUG_ENTER("ha_mroonga::store_fields_from_primary_table");
+  my_ptrdiff_t ptr_diff = PTR_BYTE_DIFF(buf, table->record[0]);
+  int i;
+  int n_columns = table->s->fields;
+  for (i = 0; i < n_columns; i++) {
+    Field *field = table->field[i];
+    if (bitmap_is_set(table->read_set, field->field_index) ||
+        bitmap_is_set(table->write_set, field->field_index)) {
+#ifndef DBUG_OFF
+      my_bitmap_map *tmp_map = dbug_tmp_use_all_columns(table,
+        table->write_set);
+#endif
+      DBUG_PRINT("info",("mroonga store column %d(%d)",i,field->field_index));
+      field->move_field_offset(ptr_diff);
+      mrn_store_field(ctx, field, col[i], rid);
+      field->move_field_offset(-ptr_diff);
+#ifndef DBUG_OFF
+      dbug_tmp_restore_column_map(table->write_set, tmp_map);
+#endif
+    }
+  }
+  DBUG_VOID_RETURN;
+}
+
 #ifdef __cplusplus
 }
 #endif
